# R FUNCTIONS TO GENERATE DATA SETS FROM A DIRICHLET DIFFUSION TREE PRIOR. 
# Also includes functions to display the latent tree structure.
#
# Written by Radford M. Neal, March 2001.
#
# These functions were used to generate the examples in the following paper:
#
#   Neal, R. M. (2001) "Defining priors for distributions using Dirichlet
#     diffusion trees".


# TREE REPRESENTATION.  A diffusion tree is represented using R lists, each
# of which corresponds to a node in the tree (which may be terminal), or
# equivalently to a segment ending at that node.  These lists contain the
# following elements:
#
#   count        The number of paths that come into this node.
#
#   data         The vector giving the location of this node, which will be
#                a data point if the node is terminal.
#
#   time         The time, from 0 to 1, for this node.  The time will be 1
#                for a terminal node.
# 
#   A            Value of the cumulative divergence function evaluated with
#                time as its argument.  Not present for terminal nodes.
#
#   left, right  The lists for the nodes that are the children of this node.
#                There is no significance to which is left and which is right.
#                These element are not present for terminal nodes.
#
# The tree as a whole is represented by the list for the first node - ie,
# for the first segment, which starts at time 0.


# GENERATE A DATA SET FROM A DIFFUSION TREE PRIOR.  Each of the n data items 
# will consist of p variables.  The function Ainv is the inverse of the 
# cumulative divergence function.  The value returned is a list containing
# the data and the tree.  Note that the sigma parameter of the Dirichlet
# diffusion tree is always taken to be one.

dft.gen <- function (p, Ainv, n)
{
  data <- matrix(0,n,p)
  data[1,] <- rnorm(p)
  tree <- list (time=1, count=1, data=data[1,])

  if (n==1) return (data,tree)

  for (i in 2:n)
  { 
    r <- dft.add.one (p, Ainv, tree)
    data[i,] <- r$data
    tree <- r$tree
  }

  return (data,tree)
}


# ADD ONE MORE DATA ITEM TO A TREE.  Generates one additional data item,
# consisting of p variables, to add to the data set whose underlying 
# tree is passed as an argument.  The function Ainv is the inverse of the
# cumulative divergence function.  The value returned is a list consisting 
# of the new data vector and the updated tree.

# The start.time, start.A, and start.pos arguments default to zero for the 
# outside use of this function.  When this function calls itself recursively,
# they are set to the starting time of the diffusion for the current subtree,
# to the cumulative divergence function evaluated at that time, and to the
# position in data space reached at that time.

dft.add.one <- function (p, Ainv, tree, 
                 start.time=0, start.A=0, start.pos=rep(0,p))
{
  # Find divergence time from initial segment.

  A <- start.A + rexp(1,1/tree$count)
  time <- Ainv(A)

  # Check for something going wrong.

  if (time<start.time)
  { stop("Somehow ended up going backwards in time!")
  }

  if (time>1)
  { stop("Inverse cumulative divergence function gave a time greater than one!")
  }

  # Special case where divergence occurs immediately.

  if (time==start.time)
  { 
    tree <- list (time=time, A=A, count=tree$count+1, data=start.pos, left=tree,
              right=list(time=1,count=1,data=rnorm(p,start.pos,sqrt(1-time))))

    return (data=tree$right$data, tree)
  }

  # Case where divergence occurs before end of segment, so new node is created.

  if (time<tree$time)
  { 
    precision <- 1/(time-start.time) + 1/(tree$time-time)
    pos <- rnorm (p, 
      (start.pos/(time-start.time) + tree$data/(tree$time-time)) / precision,
      1/sqrt(precision))
    
    tree <- list (time=time, A=A, count=tree$count+1, data=pos, left=tree,
                  right=list(time=1,count=1,data=rnorm(p,pos,sqrt(1-time))))

    return (data=tree$right$data, tree)
  }

  # Case where divergence does not occur before time one, so data is reused.

  if (tree$time==1)
  { 
    tree$count <- tree$count + 1
    return (data=tree$data, tree)
  } 

  # Case where no divergence occurs, and so we must select a branch to go down.

  if (tree$left$count+tree$right$count!=tree$count)
  { stop("Inconsistent counts!")
  }

  if (runif(1)<tree$left$count/tree$count)
  { r <- dft.add.one(p,Ainv,tree$left,tree$time,tree$A,tree$data)
    tree$left <- r$tree
    tree$count <- tree$count+1
  }
  else
  { r <- dft.add.one(p,Ainv,tree$right,tree$time,tree$A,tree$data)
    tree$right <- r$tree
    tree$count <- tree$count+1
  }

  return (data=r$data, tree)

}


# PLOT THE TREE UNDERLYING 1D DATA.  Parameters are the tree and the 
# dimension to plot (default 1), along with whether to label nodes with
# the number of paths passing through (default F), whether to plot the path
# between nodes as Brownian motion (default F), as opposed to a straight
# line, whether to plot dots at the nodes (default F), a random seed
# (relvant only if Brownian motion is being plotted), and the range
# of data values (defaults to the actual range).  The init, st.time, and
# st.pos parameters are only for internal use in recursive calls.

dft.plot1 <- function (tree, j=1, label=F, brownian=F, dots=F, seed=NULL, 
               range=NULL, init=T, st.time=0, st.pos=0)
{
  if (!is.null(seed))
  { set.seed(seed)
  } 

  if (init) 
  { if (is.null(range))
    { range <- dft.range(tree,j)
    }
    plot (c(0,0,1), c(0, range), type="n", xlab="", ylab="")
  }

  if (label)
  { points (c(st.time,tree$time), c(st.pos,tree$data[j]), type="b", pch=" ")
    if (tree$count>=10)
    { ch <- "#"
    }
    else if (tree$count>1)
    { ch <- as.character(tree$count)
    }
    else
    { ch <- 19
    }
    points (tree$time, tree$data[j], pch=ch)
  }
  else if (brownian)
  { dft.brownian (st.time, tree$time, st.pos, tree$data[j])
  }
  else
  { points (c(st.time,tree$time), c(st.pos,tree$data[j]), type="l")
  }

  if (dots)
  { points (c(st.time,tree$time), c(st.pos,tree$data[j]), type="p", pch=20)
  }

  if (!is.null(tree$left) )
  { dft.plot1(tree$left,j,label,brownian,dots,NULL,
              range,F,tree$time,tree$data[j])
  }

  if (!is.null(tree$right))
  { dft.plot1(tree$right,j,label,brownian,dots,NULL,
              range,F,tree$time,tree$data[j])
  }
}


# PLOT AN APPROXIMATION TO 1D BROWNIAN MOTION.  The plot starts at time t1 
# and location x1 and ends at time t2 and location x2.  The path in between 
# is randomly generated conditional of these endpoints.

dft.brownian <- function (t1, t2, x1, x2)
{ 
  td <- (t2-t1)/2

  if (td<0.001)
  { lines (c(t1,t2), c(x1,x2))
  }
  else
  { tm <- (t2+t1)/2  
    xm <- rnorm (1, (x1+x2)/2, sqrt(td/2))
    dft.brownian (t1, tm, x1, xm)
    dft.brownian (tm, t2, xm, x2)
  }
}


# PLOT THE TREE UNDERLYING 2D DATA.  The parameters are the tree and the
# two dimensions to plot (defaults to 1 and 2), along with whether to plot
# circles at the non-terminal node locations (default T), and the ranges
# for the two dimensions (default is to use the actual data ranges).  The
# init and st.pos parameters are for internal use when this function calls
# itself recursively.

dft.plot2 <- function (tree, j=1, k=2, circles=T, range.j=NULL, range.k=NULL, 
                        init=T, st.pos=c(0,0))
{
  if (init) 
  { if (is.null(range.j))
    { range.j <- dft.range(tree,j)
    }
    if (is.null(range.k))
    { range.k <- dft.range(tree,k)
    }
    plot (c(0,range.j), c(0,range.k), type="n", xlab="", ylab="")
  }

  if (circles)
  { points(c(st.pos[1],tree$data[j]),c(st.pos[2],tree$data[k]),type="b",pch=1)
  }
  else
  { lines(c(st.pos[1],tree$data[j]),c(st.pos[2],tree$data[k]))
  }

  if (tree$time==1)
  { points(tree$data[j],tree$data[k],type="p",pch=19)
  }

  if (!is.null(tree$left) )
  { dft.plot2(tree$left,j,k,circles,init=F,st.pos=tree$data[c(j,k)])
  }

  if (!is.null(tree$right))
  { dft.plot2(tree$right,j,k,circles,init=F,st.pos=tree$data[c(j,k)])
  }
}


# FIND THE RANGE OF DATA VALUES IN A DIFFUSION TREE.  The parameters are 
# the tree and the index of the dimension to find the range for (default 1).
# The value is the range, as a vector of low and high elements.

dft.range <- function (tree, j=1)
{ 
  low <- high <- tree$data[j]

  if (!is.null(tree$left))
  { r <- dft.range(tree$left,j)
    low <- min(low,r[1])
    high <- max(high,r[2])
  }

  if (!is.null(tree$right))
  { r <- dft.range(tree$right,j)
    low <- min(low,r[1])
    high <- max(high,r[2])
  }

  c(low,high)
}


# EXAMPLES OF INVERSE CUMULATIVE DIVERGENCE FUNCTIONS.  These won't
# work in S, only in R, since they utilize R's different scoping rules.


# Divergence function a(t) = c/(1-t).

Ainv1 <- function (c) 
{ 
  function(e) 1-exp(-e/c)  
}


# Divergence function a(t) = b + c/(1-t)^2.

Ainv2 <- function (b,c) 
{ 
  if (b==0)
  { function(e) 1 - c/(e+c)
  }
  else
  { function(e) (b + c + e - sqrt((b+c+e)^2-4*b*e)) / (2*b)
  }
}


# Equivalent of a Dirichlet process mixture with a mass parameter of a,
# with variance of the components means of v, and with variance within
# a component of 1-v.

Ainv.DPM <- function (a,v) 
{
  function (e) (e>log(1+a))*v
}
